# Alea version 1.0.0
# DiscreteEventSystems.py
# by Antoine Rauzy
# Copyrights (c) 2020 AltaRica Association
#
# This file is part of Alea, a pedagogical implementation of stochastic discrete event systems.
# Alea is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# Alea is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Alea.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1. Required modules
# 2. Discrete Event Systems
# 3. Transitions
# 4. States (to be derived)
# 5. Delays
#   5.1. Base Delays
#   5.2. Dirac Delays
#   5.3. Uniformly Distributed Delays
#   5.4. Exponentially Distributed Delays
#   5.5. Triangularly Distributed Delays
#   5.6. Stepwise Empirical Delays
#   5.7. Empirical Delays with Linear Interpolation
# 6. Weights
#   6.1. Base Weights
#   6.2. Constant Weights

# 1. Required modules
# -------------------

import sys
import math
import random

# 2. Discrete Event Systems
# -------------------------

class DiscreteEventSystem:
  def __init__(self, stateCreator):
    # REQUIRE:
    # - stateCreator is a function that takes the system as parameter and returns a state
    self.stateCreator = stateCreator
    self.transitions = dict()

  def NewState(self):
    return self.stateCreator(self)

  def GetTransitions(self):
    return self.transitions.values()

  def LookForTransition(self, name):
    return self.transitions.get(name, None)

  def NewTransition(self, name, guard, action, delay, weight = None):
    # REQUIRE:
    # - name is a string
    # - guard is a function that takes a state as parameter and returns a Boolean
    # - action is a function that takes a state as parameter and modifies it
    # - delay is a Delay object
    # - weight is a Weight object
    transition = Transition(self, name, guard, action, delay, weight)
    self.transitions[name] = transition
    return transition

  def DeleteTransition(self, transition):
    if self.LookForTransition(transition.GetName())!=transition:
      return
    self.transitions.pop(transition.GetName())

  def IsTransitionEnabled(self, transition, state):
    return transition.IsEnabled(state)

  def CalculateDelayOfTransition(self, transition, state):
    return transition.CalculateDelay(state)

  def FireTransition(self, transition, state):
    return transition.Fire(state)

# 3. Transitions
# --------------

class Transition:
  RESTART = 0
  MEMORY = 1

  def __init__(self, system, name, guard, action, delay, weight = None):
    self.system = system
    self.name = name
    self.guard = guard
    self.action = action
    self.delay = delay
    if weight==None:
      self.weight = ConstantWeight(1.0)
    else:
      self.weight = weight
    self.policy = Transition.RESTART

  def GetSystem(self):
    return self.system

  def GetName(self):
    return self.name

  def GetGuard(self):
    return self.guard

  def GetDelay(self):
    return self.delay

  def GetAction(self):
    return self.action

  def IsEnabled(self, state):
    return self.guard(state)

  def CalculateDelay(self, state):
    return self.delay.Calculate(state)

  def CalculateWeight(self, state):
    return self.weight.Calculate(state)

  def Fire(self, state):
    self.action(state)

  def GetPolicy(self):
    return self.policy

  def SetPolicy(self, policy):
    self.policy = policy


# 4. States
# ---------
# To be derived

class State:
  def __init__(self, system):
    self.system = system

  def Reset(self):
    # REQUIRE:
    # - Resets the state on the initial state of the system
    pass

# 5. Delays
# ---------

# 5.1. Base Delays
# ----------------

class Delay:
  INFINITY = sys.float_info.max

  def Calculate(self, state):
    return Delay.INFINITY

# 5.2. Dirac Delays
# -----------------

class DiracDelay(Delay):
  def __init__(self, value):
    self.value = value

  def GetValue(self):
    return self.value

  def SetValue(self, value):
    self.value = value

  def Calculate(self, state):
    return self.value

# 5.3. Uniformly Distributed Delays
# ---------------------------------

class UniformDelay(Delay):
  def __init__(self, lowValue, highValue):
    self.lowValue = lowValue
    self.highValue = highValue

  def GetLowValue(self):
    return self.lowValue

  def SetLowValue(self, value):
    self.lowValue = value

  def GetHighValue(self):
    return self.highValue

  def SetHighValue(self, value):
    self.highValue = value

  def Calculate(self, state):
    return random.uniform(self.lowValue, self.highValue)

# 5.4. Exponentially Distributed Delays
# -------------------------------------

class ExponentialDelay(Delay):
  def __init__(self, mean):
    self.mean = mean

  def GetMean(self):
    return self.rate

  def SetRate(self, mean):
    self.mean = mean

  def Calculate(self, state):
    z = random.random()
    return -math.log(z)*self.mean

# 5.5. Triangularly Distributed Delays
# ------------------------------------

class TriangularDelay(Delay):
  def __init__(self, lowValue, highValue, mode):
    self.lowValue = lowValue
    self.highValue = highValue
    self.mode = mode

  def GetLowValue(self):
    return self.lowValue

  def SetLowValue(self, value):
    self.lowValue = value

  def GetHighValue(self):
    return self.highValue

  def SetHighValue(self, value):
    self.highValue = value

  def GetMode(self):
    return self.mode

  def SetMode(self, value):
    self.mode = value

  def Calculate(self, state):
    z = random.random()
    r = (self.mode - self.lowValue) / (self.highValue - self.lowValue)
    if z<r:
      delay = self.lowValue + math.sqrt((self.highValue - self.lowValue) * (self.mode - self.lowValue) * z)
    else:
      delay = self.highValue - math.sqrt((self.highValue - self.lowValue) * (self.highValue - self.mode) * (1 - z))
    return delay


# 5.6. Stepwise Empirical Delays
# ------------------------------

class StepwiseEmpiricalDelay(Delay):
  def __init__(self, points):
    # REQUIRE: points must be a list of pairs representing a cumulative probability distribution
    self.points = points

  def GetPoints(self):
    return self.points

  def SetPoints(self, points):
    self.points = points

  def Calculate(self, state):
    z = random.random()
    delay = 0
    for (x, y) in self.points:
      if y>z:
        return delay
      delay = x
    return Delay.INFINITY

# 5.7. Empirical Delays with Linear Interpolation
# -----------------------------------------------

class LinearEmpiricalDelay(Delay):
  def __init__(self, points):
    # REQUIRE: points must be a list of pairs representing a cumulative probability distribution
    self.points = points

  def GetPoints(self):
    return self.points

  def SetPoints(self, points):
    self.points = points

  def Calculate(self, state):
    z = random.random()
    x0 = 0
    y0 = 0
    for (x1, y1) in self.points:
      if y1>z:
        return x0 + (x1 - x0) * (z - y0) / (y1 - y0)
      x0 = x1
      y0 = y1
    return Delay.INFINITY

# 6. Weights
# ----------

# 6.1. Base Weights
# -----------------

class Weight:
  def Calculate(self, state):
    return 1.0

# 6.2. Constant Weights
# ---------------------

class ConstantWeight(Weight):
  def __init__(self, weight):
    self.weight = weight

  def Calculate(self, state):
    return self.weight

