# TPK4186 - Advanced Tools for Performance Engineering
# IsingModel.py
# by Antoine Rauzy
# Copyrights (c) 2019-2020 NTNU
#
# This file is part of the educational material provided to NTNU students registered to the course TPK4186.
# It is licensed under a Creative Commons Attribution-NonCommercial 4.0 International License.
# For details about this license, see https://creativecommons.org/licenses/by-nc/4.0/.

# Table of Contents
# -----------------
# 1. Required modules
# 2. Maps
# 3. Animators
# 4. Main

# To get this program work, you need run the following command in the Spyder console
# %matplotlib qt

# 1. Required modules
# -------------------

import random
import math
import numpy as np
import matplotlib.pyplot as plt
from matplotlib import colors
import matplotlib.animation as animation

# 2. Maps
# -------

class Map:
  UP = 0.25
  DOWN = 0.75

  def __init__(self, numberOfRows, numberOfColumns):
    self.numberOfRows = numberOfRows 
    self.numberOfColumns = numberOfColumns
    self.grids = np.zeros((2, numberOfRows, numberOfColumns))
    self.currentGridIndex = 0
    self.upProbability = 0.5
    self.temperature = 2

  def GetCurrentGrid(self):
    return self.grids[self.currentGridIndex]

  def GetOppositeState(self, state):
    if state==Map.UP:
      return Map.DOWN
    return Map.UP

  def GetState(self, row, column):
    return self.grids[self.currentGridIndex][row][column]

  def SetState(self, row, column, state):
    self.grids[self.currentGridIndex][row][column] = state

  def CountStatesOfNeighborhood(self, row, column, state):
    count = 0
#    if self.GetState(row, column)==state:
#      count += 1
    if row>0 and self.GetState(row-1, column)==state:
      count += 1
    if column>0 and self.GetState(row, column-1)==state:
      count += 1
    if row<self.numberOfRows-1 and self.GetState(row+1, column)==state:
      count += 1
    if column<self.numberOfColumns-1 and self.GetState(row, column+1)==state:
      count += 1
    return count

  def Initialize(self):
    for row in range(0, self.numberOfRows):
      for column in range(0, self.numberOfColumns):
        if random.random()<=self.upProbability:
          self.SetState(row, column, Map.UP)
        else:
          self.SetState(row, column, Map.DOWN)

  def UpdateState(self):
    nextGridIndex = 1 - self.currentGridIndex
    for row in range(0, self.numberOfRows):
      for column in range(0, self.numberOfColumns):
        state = self.GetState(row, column)
        deltaEnergy = 2 * self.CountStatesOfNeighborhood(row, column, state)
        transitionProbability = math.exp(-deltaEnergy/self.temperature)
        if random.random()<=transitionProbability:
          self.grids[nextGridIndex][row][column] = self.GetOppositeState(state)
        else:
          self.grids[nextGridIndex][row][column] = state
    self.currentGridIndex = nextGridIndex

# 3. Animators
# ------------

class Animator:
  TheAnimator = None

  def __init__(self, map, width, height):
    self.map = map
    colorMap = colors.ListedColormap(['white', 'black'])
    bounds = [0, 0.5, 1.0]
    norm = colors.BoundaryNorm(bounds, colorMap.N)
    self.figure = plt.figure(figsize=(width, height))
    self.image = plt.imshow(self.map.GetCurrentGrid(), cmap=colorMap, norm=norm, aspect='equal')
    self.animation = None
    self.maximumNumberOfSteps = 0
    Animator.TheAnimator = self

  def ShowCurrentMap(self):
    self.image.set_array(self.map.GetCurrentGrid())
    # plt.savefig("Ising0.pdf")
    plt.show()

  @staticmethod
  def ShowNextMap(i):
    self = Animator.TheAnimator
    self.map.UpdateState()
    self.image.set_array(self.map.GetCurrentGrid())
    # plt.savefig("step{0:d}.pdf".format(self.map.GetNumberOfStepsBeforeExtinction()))
    return self.image,

  @staticmethod
  def Iterator():
    self = Animator.TheAnimator
    numberOfSteps = 0
    while numberOfSteps<self.maximumNumberOfSteps:
      numberOfSteps += 1
      # print("{0:d}\t{1:d}".format(numberOfSteps, self.map.GetNumberOfLivingCells()))
      yield numberOfSteps

  def LaunchAnimation(self, maximumNumberOfSteps, timeStep):
    self.maximumNumberOfSteps = maximumNumberOfSteps
    self.image.set_array(self.map.GetCurrentGrid())
    # plt.savefig("step0.pdf")
    self.animation = animation.FuncAnimation(self.figure, Animator.ShowNextMap, frames = Animator.Iterator, interval = timeStep, repeat = False)
    plt.show()

# 5. Main
# -------

map = Map(100, 100)
map.upProbability = 0.5
map.temperature = 2.0

map.Initialize()


animator = Animator(map, 10, 10)
#map.UpdateState()
#map.UpdateState()
#animator.ShowCurrentMap()
animator.LaunchAnimation(1000, 100)


