# TPK4186 - Advanced Tools for Performance Engineering
# SchellingModel.py
# by Antoine Rauzy
# Copyrights (c) 2019-2020 NTNU
#
# This file is part of the educational material provided to NTNU students registered to the course TPK4186.
# It is licensed under a Creative Commons Attribution-NonCommercial 4.0 International License.
# For details about this license, see https://creativecommons.org/licenses/by-nc/4.0/.

# Table of Contents
# -----------------
# 1. Required modules
# 2. Maps
# 3. Animators
# 4. Main

# To get this program work, you need run the following command in the Spyder console
# %matplotlib qt

# 1. Required modules
# -------------------

import random
import math
import numpy as np
import matplotlib.pyplot as plt
from matplotlib import colors
import matplotlib.animation as animation

# 2. Maps
# -------

class Map:
  WHITE = 0.40
  BLACK = 0.80
  EMPTY = 0.95

  def __init__(self, numberOfRows, numberOfColumns):
    self.numberOfRows = numberOfRows 
    self.numberOfColumns = numberOfColumns
    self.grid = np.zeros((numberOfRows, numberOfColumns))
    self.whiteProbability = 0.45
    self.blackProbability = 0.45
    self.happinessRatio = 0.3
    self.emptyHouses = []
    self.unhappyFamilies = []

  def GetCommunity(self, row, column):
    return self.grid[row][column]

  def SetCommunity(self, row, column, community):
    self.grid[row][column] = community

  def GetGrid(self):
    return self.grid

  def GetNumberOfUnhappyFamilies(self):
    return len(self.unhappyFamilies)

  def Initialize(self):
    self.emptyHouses.clear()
    for row in range(0, self.numberOfRows):
      for column in range(0, self.numberOfColumns):
        z = random.random()
        if z<=self.whiteProbability:
          self.SetCommunity(row, column, Map.WHITE)
        elif z<=self.whiteProbability+self.blackProbability:
          self.SetCommunity(row, column, Map.BLACK)
        else:
          self.SetCommunity(row, column, Map.EMPTY)
          self.emptyHouses.append([row, column])
    self.UpdateUnhappyFamilyList()

  def GetCommunityRatioInNeighborhood(self, row, column, community):
    numberOfNeighbors = 0
    numberOfNeighborsOfTheSameCommunity = 0
    for r in range(row-2, row+3):
      if r<0 or r>=self.numberOfRows:
        continue
      for c in range(column-2, column+3):
        if c<0 or c>=self.numberOfColumns:
          continue
        numberOfNeighbors += 1
        if self.GetCommunity(r, c)==community:
          numberOfNeighborsOfTheSameCommunity += 1
    return numberOfNeighborsOfTheSameCommunity/numberOfNeighbors

  def UpdateUnhappyFamilyList(self):
    self.unhappyFamilies.clear()
    for row in range(0, self.numberOfRows):
      for column in range(0, self.numberOfColumns):
        community = self.GetCommunity(row, column)
        if community!=Map.EMPTY:
          ratio = self.GetCommunityRatioInNeighborhood(row, column, community)
          if ratio<self.happinessRatio:
            self.unhappyFamilies.append((row, column))
    if len(self.unhappyFamilies)==0:
      print("No more unhappy family")
      return

  def UpdateState(self):
    random.shuffle(self.unhappyFamilies)
    for (row, column) in self.unhappyFamilies:
      community = self.GetCommunity(row, column)
      emptyHouseIndex = random.randrange(0, len(self.emptyHouses))
      emptyHouse = self.emptyHouses[emptyHouseIndex]
      emptyHouseRow = emptyHouse[0]
      emptyHouseColumn = emptyHouse[1]
      self.SetCommunity(emptyHouseRow, emptyHouseColumn, community)
      self.SetCommunity(row, column, Map.EMPTY)
      emptyHouse[0] = row
      emptyHouse[1] = column

# 3. Animators
# ------------

class Animator:
  TheAnimator = None

  def __init__(self, map, width, height):
    self.map = map
    colorMap = colors.ListedColormap(['white', 'black', 'grey'])
    bounds = [0, 0.45, 0.90, 1.0]
    norm = colors.BoundaryNorm(bounds, colorMap.N)
    self.figure = plt.figure(figsize=(width, height))
    self.image = plt.imshow(self.map.GetGrid(), cmap=colorMap, norm=norm, aspect='equal')
    self.animation = None
    self.maximumNumberOfSteps = 0
    Animator.TheAnimator = self

  def ShowCurrentMap(self):
    self.image.set_array(self.map.GetGrid())
    # plt.savefig("Ising0.pdf")
    plt.show()

  @staticmethod
  def ShowNextMap(i):
    self = Animator.TheAnimator
    self.map.UpdateState()
    self.image.set_array(self.map.GetGrid())
    # plt.savefig("step{0:d}.pdf".format(self.map.GetNumberOfStepsBeforeExtinction()))
    return self.image,

  @staticmethod
  def Iterator():
    self = Animator.TheAnimator
    numberOfSteps = 0
    while numberOfSteps<self.maximumNumberOfSteps:
      numberOfSteps += 1
      self.map.UpdateUnhappyFamilyList()
      if self.map.GetNumberOfUnhappyFamilies()==0:
        break
      # print("{0:d}\t{1:d}".format(numberOfSteps, self.map.GetNumberOfLivingCells()))
      yield numberOfSteps

  def LaunchAnimation(self, maximumNumberOfSteps, timeStep):
    self.maximumNumberOfSteps = maximumNumberOfSteps
    self.image.set_array(self.map.GetGrid())
    # plt.savefig("step0.pdf")
    self.animation = animation.FuncAnimation(self.figure, Animator.ShowNextMap, frames = Animator.Iterator, interval = timeStep, repeat = False)
    plt.show()

# 5. Main
# -------

random.seed(12345)
map = Map(100, 100)
map.happinessRatio = 0.34

map.Initialize()


animator = Animator(map, 10, 10)
#map.UpdateState()
#map.UpdateState()
#animator.ShowCurrentMap()
animator.LaunchAnimation(1000, 100)


